import pandas as pd
import numpy as np

def _signal_features(window_data_df, channels, fs):
    features = pd.DataFrame()

    # Compute signal entropy
    feat_df_signal_entropy = _signal_entropy(window_data_df, channels)

    # Compute RMS
    feat_df_signal_rms = _signal_rms(window_data_df, channels)

    # Compute range
    feat_df_signal_range = _signal_range(window_data_df, channels)

    # Compute IQR of Autocovariance
    feat_df_iqr_auto = _iqr_of_autocovariance(window_data_df, channels)

    # Compute Dominant Frequency
    sampling_rate = fs
    frequncy_cutoff = 12.0
    feat_df_dom_freq = _dominant_frequency(window_data_df, sampling_rate, frequncy_cutoff, channels)

    # Compute mean cross rate
    feat_df_mean_cross_rate = _mean_cross_rate(window_data_df, channels)

    features = features.join(feat_df_signal_entropy, how='outer')
    features = features.join(feat_df_signal_rms, how='outer')
    features = features.join(feat_df_signal_range, how='outer')
    features = features.join(feat_df_iqr_auto, how='outer')
    features = features.join(feat_df_dom_freq, how='outer')
    features = features.join(feat_df_mean_cross_rate, how='outer')

    return features

def _signal_entropy(signal_df, channels):
    signal_entropy_df = pd.DataFrame()

    for channel in channels:
        data_norm = signal_df[channel]/np.std(signal_df[channel])
        h, d = _histogram(data_norm)

        lowerbound = d[0]
        upperbound = d[1]
        ncell = int(d[2])

        estimate = 0
        sigma = 0
        count = 0

        for n in range(ncell):
            if h[n] != 0:
                logf = np.log(h[n])
            else:
                logf = 0
            count = count + h[n]
            estimate = estimate - h[n] * logf
            sigma = sigma + h[n] * logf ** 2

        nbias = -(float(ncell) - 1) / (2 * count)

        estimate = estimate / count
        estimate = estimate + np.log(count) + np.log((upperbound - lowerbound) / ncell) - nbias

        # Scale the entropy estimate to stretch the range
        estimate = np.exp(estimate ** 2) - np.exp(0) - 1

        signal_entropy_df[channel + '_signal_entropy'] = [estimate]

    return signal_entropy_df

def _signal_rms(signal_df, channels):
    rms_df = pd.DataFrame()

    for channel in channels:
        rms_df[channel + '_rms'] = [np.std(signal_df[channel] - signal_df[channel].mean())]

    return rms_df

def _signal_range(signal_df, channels):
    range_df = pd.DataFrame()

    for channel in channels:
        range_df[channel + '_range'] = [signal_df[channel].max(skipna=True) - signal_df[channel].min(skipna=True)]

    return range_df

def _iqr_of_autocovariance(signal_df, channels):
    from scipy import stats
    from statsmodels.tsa.stattools import acf

    autocov_range_df = pd.DataFrame()
    n_samples = signal_df.shape[0]
    for channel in channels:
        current_autocov_iqr = stats.iqr(acf(signal_df[channel], unbiased=True, nlags=n_samples/2))
        autocov_range_df[channel + '_iqr_of_autocovariance'] = [current_autocov_iqr]

    return autocov_range_df

def _dominant_frequency(signal_df, sampling_rate, cutoff, channels):
    from scipy import stats

    dominant_freq_df = pd.DataFrame()
    for channel in channels:
        signal_x = signal_df[channel]

        padfactor = 1
        dim = signal_x.shape
        nfft = 2 ** ((dim[0] * padfactor).bit_length())

        freq_hat = np.fft.fftfreq(nfft) * sampling_rate
        freq = freq_hat[0:nfft / 2]

        idx1 = freq <= cutoff
        idx_cutoff = np.argwhere(idx1)
        freq = freq[idx_cutoff]

        sp_hat = np.fft.fft(signal_x, nfft)
        sp = sp_hat[0:nfft / 2] * np.conjugate(sp_hat[0:nfft / 2])
        sp = sp[idx_cutoff]
        sp_norm = sp / sum(sp)

        max_freq = freq[sp_norm.argmax()][0]
        max_freq_val = sp_norm.max().real

        idx2 = (freq > max_freq - 0.5) * (freq < max_freq + 0.5)
        idx_freq_range = np.where(idx2)[0]
        dom_freq_ratio = sp_norm[idx_freq_range].real.sum()

        # Calculate spectral flatness
        spectral_flatness = 10.0*np.log10(stats.mstats.gmean(sp_norm)/np.mean(sp_norm))

        # Estimate spectral entropy
        spectral_entropy_estimate = 0
        for isess in range(len(sp_norm)):
            if sp_norm[isess] != 0:
                logps = np.log2(sp_norm[isess])
            else:
                logps = 0
            spectral_entropy_estimate = spectral_entropy_estimate - logps * sp_norm[isess]

        spectral_entropy_estimate = spectral_entropy_estimate / np.log2(len(sp_norm))
        # spectral_entropy_estimate = (spectral_entropy_estimate - 0.5) / (1.5 - spectral_entropy_estimate)

        dominant_freq_df[channel + '_dom_freq_value'] = [max_freq]
        dominant_freq_df[channel + '_dom_freq_magnitude'] = [max_freq_val]
        dominant_freq_df[channel + '_dom_freq_ratio'] = [dom_freq_ratio]
        dominant_freq_df[channel + '_spectral_flatness'] = [spectral_flatness[0].real]
        dominant_freq_df[channel + '_spectral_entropy'] = [spectral_entropy_estimate[0].real]

    return dominant_freq_df

def _mean_cross_rate(signal_df, channels):
    '''
    Compute mean cross rate of sensor signals.

    :param signal_df: dataframe housing desired sensor signals
    :param channels: channels of signal to measure mean cross rate
    :return: dataframe housing calculated mean cross rate for each signal channel
    '''
    mean_cross_rate_df = pd.DataFrame()
    signal_df_mean = signal_df[channels] - signal_df[channels].mean()

    for channel in channels:
        MCR = 0

        for i in range(len(signal_df_mean) - 1):
            if np.sign(signal_df_mean.loc[i, channel]) != np.sign(signal_df_mean.loc[i + 1, channel]):
                MCR += 1

        MCR = float(MCR) / len(signal_df_mean)

        mean_cross_rate_df[channel + '_mean_cross_rate'] = [MCR]

    return mean_cross_rate_df

def _range_count_percentage(signal_df, channels, min_value=-1, max_value=1):
    import tsfresh as tsf

    range_count_df = pd.DataFrame()
    for channel in channels:
        signal_x = signal_df[channel]
        current_range_count = tsf.feature_extraction.feature_calculators.range_count(signal_x, min_value, max_value) * 1.0 / len(signal_x)
        range_count_df[channel + '_range_count_per'] = [current_range_count]

    return range_count_df

def _histogram(signal_x):
    descriptor = np.zeros(3)

    ncell = np.ceil(np.sqrt(len(signal_x)))

    max_val = np.nanmax(signal_x.values)
    min_val = np.nanmin(signal_x.values)

    delta = (max_val - min_val) / (len(signal_x) - 1)

    descriptor[0] = min_val - delta / 2
    descriptor[1] = max_val + delta / 2
    descriptor[2] = ncell

    h = np.histogram(signal_x, ncell.astype(int), range=(min_val, max_val))

    return h[0], descriptor


class Gaitpy():
    ''' Gait feature extraction and bout classification from single accelerometer in the lumbar location

    Using vertical acceleration data from the lumbar location, this class includes functions for:
        - Continuous wavelet based method of gait kinematic feature extraction.
        - Machine learning based method of bout classification.

    Parameters
    ----------
    data : str or pandas.core.frame.DataFrame
        Pandas dataframe containing timestamp column and vertical acceleration data during gait, both of type float
        OR
        File path of .csv file containing timestamp column and vertical acceleration data during gait
            - One column should contain unix timestamps of type float -- by default gaitpy will assume the column title is
                                                                         'timestamps' with units in milliseconds
            - A second column should be vertical acceleration of type float -- by default gaitpy will assume the column
                                                                               title is 'y' with units in m/s^2

    sample_rate : int or float
        Sampling rate of accelerometer data in Hertz.

    v_acc_col_name : str
        Column name of the vertical acceleration data ('y' by default)

    ts_col_name : str
        Column name of the timestamps ('timestamps' by default)

    v_acc_units : str
        Units of vertical acceleration data ('m/s^2' by default)
        Options:
            - 'm/s^2' = meters per second squared
            - 'g' = standard gravity

    ts_units : str
        Units of timestamps ('ms' by default)
        Options:
            - 's' = seconds
            - 'ms' = milli-seconds
            - 'us' = microseconds

    flip : bool
        Boolean specifying whether to flip vertical acceleration data before analysis (False by default). Algorithm
        assumes that baseline vertical acceleration data is at -9.8 m/s^2 or -1g. (ie. If baseline data in vertical
        direction is 1g, set 'flip' argument to True)

    down_sample : int or float
        Sampling rate to downsample data to. Helps to standardize results from multiple sensors with different
        sampling rates.
    '''

    def __init__(self, data, sample_rate, v_acc_col_name='y', ts_col_name='timestamps', v_acc_units='m/s^2', ts_units='ms', flip=False, down_sample=50):
        self.data = data
        self.sample_rate = sample_rate
        self.v_acc_col_name = v_acc_col_name
        self.ts_col_name = ts_col_name
        self.v_acc_units = v_acc_units
        self.ts_units = ts_units
        self.flip = flip
        self.down_sample = down_sample

    def extract_features(self, subject_height, subject_height_units='centimeter', sensor_height_ratio=0.53, result_file=None, classified_gait=None, ic_prom=5, fc_prom=25):
        ''' Continuous wavelet transform based method of gait feature detection optimization methods

        Parameters
        ----------
        subject_height : int or float
            Height of the subject. Accepts centimeters by default.

        subject_height_units : str
            Units of provided subject height. Centimeters by default.
            - options: 'centimeter', 'inches', 'meter'

        sensor_height_ratio : float
            Height of the sensor relative to subject height. Calculated: sensor height / subject height

        result_file : str
            Optional argument that accepts .csv filepath string to save resulting gait feature dataframe to.
            None by default. (ie. myfolder/myfile.csv)

        classified_gait : str or pandas.core.frame.DataFrame
            Pandas dataframe containing results of gait bout classification procedure (classify_bouts)
            OR
            File path of .h5 file containing results of gait bout classification procedure (classify_bouts)

        ic_prom : int
            Prominance of initial contact peak detection

        fc_prom : int
            Prominance of final contact peak detection

        '''
        import pandas as pd
        import util
        import warnings

        print('\tExtracting features...')

        # Load data
        y_accel, timestamps = util._load_data(self, self.down_sample)

        # If classified gait is provided, load pandas dataframe or h5 file
        if classified_gait is not None:
            if type(classified_gait) is str:
                gait_predictions = pd.read_hdf(classified_gait)
            elif type(classified_gait) is pd.core.frame.DataFrame:
                gait_predictions = classified_gait
            else:
                print('Unable to load classified gait: Please make sure the data is in the correct format, aborting...')
                return
            # Isolate gait bouts
            gait_windows = gait_predictions[gait_predictions['prediction'] == 1]
            if gait_windows.empty:
                print('The classified_gait data indicates no bouts of gait were detected, aborting...')
                return

            # Concatenate concurrent bouts
            gait_bouts = util._concatenate_windows(gait_windows, window_length=3)
        else:
            # if classified_gait is not provided, assume entire timeseries is 1 bout of gait
            start_time = timestamps[0].astype('datetime64[ms]')
            end_time = timestamps.iloc[-1].astype('datetime64[ms]')
            gait_bouts = pd.DataFrame(data={'start_time': [start_time],
                                            'end_time': [end_time],
                                            'bout_length': [(end_time - start_time).item().total_seconds()]})

        all_bout_gait_features = pd.DataFrame()
        bout_n = 1
        # Loop through gait bouts
        for row_n, bout in gait_bouts.iterrows():
            bout_indices = (timestamps.astype('datetime64[ms]') >= bout.start_time) & (timestamps.astype('datetime64[ms]') <= bout.end_time)
            bout_data = pd.DataFrame([])
            bout_data['y'] = pd.DataFrame(y_accel.loc[bout_indices].reset_index(drop=True))
            bout_data['ts'] = timestamps.loc[bout_indices].reset_index(drop=True)
            if len(bout_data.y) < 15:
                warnings.warn('There are too few data points between '+str(bout.start_time)+' and '+str(bout.end_time)+', skipping bout...')
                continue

            # Run CWT Gait Model IC and FC detection
            ic_peaks, fc_peaks = util._cwt(bout_data.y, self.down_sample, ic_prom, fc_prom)

            # Run gait cycle optimization procedure
            pd.options.mode.chained_assignment = None
            optimized_gait = util._optimization(bout_data['ts'], ic_peaks, fc_peaks)
            if optimized_gait.empty or 1 not in list(optimized_gait.Gait_Cycle):
                continue

            # Calculate changes in height of the center of mass
            optimized_gait = util._height_change_com(optimized_gait, bout_data['ts'], bout_data['y'], self.down_sample)

            # Calculate gait features
            sensor_height = util._calculate_sensor_height(subject_height, subject_height_units, sensor_height_ratio)
            gait_features = util._cwt_feature_extraction(optimized_gait, sensor_height)

            # remove center of mass height and gait cycle boolean columns, remove rows with NAs
            gait_features.dropna(inplace=True)
            gait_features.drop(['CoM_height','Gait_Cycle', 'FC_opp_foot'], axis=1, inplace=True)

            gait_features.insert(0, 'bout_number', bout_n)
            gait_features.insert(1, 'bout_length_sec', bout.bout_length)
            gait_features.insert(2, 'bout_start_time', bout.start_time)
            gait_features.insert(5, 'gait_cycles', len(gait_features))
            all_bout_gait_features = all_bout_gait_features.append(gait_features)

            bout_n += 1
        all_bout_gait_features.reset_index(drop=True, inplace=True)
        all_bout_gait_features.iloc[:,7:] = all_bout_gait_features.iloc[:,7:].round(3)

        # Save results
        if result_file:
            try:
                if not result_file.endswith('.csv'):
                    result_file += '.csv'
                all_bout_gait_features.to_csv(result_file, index=False, float_format='%.3f')
            except:
                print('Unable to save data: Please make sure your results directory exists, aborting...')
                return

        if all_bout_gait_features.empty:
            print('\tFeature extraction complete. No gait cycles detected...\n')
        else:
            print('\tFeature extraction complete!\n')

        return all_bout_gait_features

    def plot_contacts(self, gait_features, result_file=None, show_plot=True):
        """ Plot initial and final contacts of lumbar based gait feature extraction

        Parameters
        ----------
        gait_features : pandas.DataFrame or str
            Pandas dataframe containing results of extract_features function
            OR
            File path of .csv file containing results of extract_features function

        result_file : str
            Optional argument that accepts .html filepath string to save resulting gait event plot to.
            None by default. (ie. myfolder/myfile.html)

        show_plot : bool
            Optional boolean argument that specifies whether your plot is displayed. True by default.

        """
        from bokeh.plotting import figure, output_file, save, show
        from bokeh.models import Legend, Span
        import pandas as pd
        import util
        import numpy as np

        print('\tPlotting contacts...')

        # Load data
        y_accel, timestamps = util._load_data(self, self.down_sample)
        ts = pd.to_datetime(timestamps, unit='ms')

        # Load gait_features
        try:
            if type(gait_features) is str:
                icfc = pd.read_csv(gait_features)
            elif type(gait_features) is pd.core.frame.DataFrame:
                icfc = gait_features
            else:
                print('Unable to load gait features: Please make sure the gait_features is in the correct format, aborting...')
                return
        except:
            print('Unable to load gait features: Please make sure you have provided the correct filepath or dataframe, aborting...')
            return

        if icfc.empty:
            print('\tGait feature dataframe is empty, aborting...')
            return

        p = figure(plot_width=1200, plot_height=600, x_axis_label='Time', y_axis_label='m/s^2', toolbar_location='above', x_axis_type='datetime')
        # Plot vertical axis
        p1 = p.line(ts, y_accel, line_width=2, line_color='blue')

        # isolate ICs, FCs, and bout start/end times
        minima_time = []
        minima_signal = []
        maxima_time = []
        maxima_signal = []
        bout_starts = []
        bout_ends = []
        ics = pd.to_datetime(icfc.IC, unit='ms')
        fcs = pd.to_datetime(icfc.FC, unit='ms')
        icfc.bout_start_time = icfc.bout_start_time.astype(np.int64).values // 10 ** 6
        bouts = icfc[['bout_number', 'bout_length_sec', 'bout_start_time']].drop_duplicates()
        for ic in ics:
            minima_time.append(ic)
            minima_signal.append(float(y_accel[ts.index[ts == ic]]))
        for fc in fcs:
            maxima_time.append(fc)
            maxima_signal.append(float(y_accel[ts.index[ts == fc]]))
        for row, bout in bouts.iterrows():
            bout_starts.append(bout.bout_start_time)
            bout_ends.append(bout.bout_start_time + (bout.bout_length_sec*1000))

        # add IC and FCs to plot
        p2 = p.circle(minima_time, minima_signal, size=15, color="green", alpha=0.5)
        p3 = p.circle(maxima_time, maxima_signal, size=15, color="darkorange", alpha=0.5)

        # add bout start and end times to plot
        for bout_start in bout_starts:
            start_bout_line = Span(location=bout_start,
                                   dimension='height', line_color='green',
                                   line_dash='solid', line_width=1.5)
            p.add_layout(start_bout_line)
        for bout_end in bout_ends:
            end_bout_line = Span(location=bout_end,
                                 dimension='height', line_color='red',
                                 line_dash='solid', line_width=1.5)
            p.add_layout(end_bout_line)

        # add legend
        legend = Legend(items=[
            ("Acceleration", [p1]),
            ("Initial contact", [p2]),
            ("Final contact", [p3])
        ], location=(10, 300))

        # format plot
        p.add_layout(legend, 'right')
        p.xaxis.axis_label_text_font_size = "16pt"
        p.yaxis.axis_label_text_font_size = "16pt"
        p.axis.major_label_text_font_size = '16pt'
        p.title.align = 'center'
        p.title.text_font_size = '16pt'
        p.xaxis.axis_label_text_font_style = 'normal'
        p.yaxis.axis_label_text_font_style = 'normal'
        p.xaxis.axis_label_standoff = 5
        p.yaxis.axis_label_standoff = 20
        p.legend.label_text_font = "arial"
        p.legend.label_text_font_size = '16pt'
        p.legend.glyph_height = 30

        if show_plot:
            show(p)

        # save plot
        if result_file:
            try:
                if not result_file.endswith('.html'):
                    result_file += '.html'
                output_file(result_file)
                save(p)
            except:
                print('Unable to save data: Please make sure your results directory exists, aborting...')
                return

        print('\tPlot complete!\n')

    def classify_bouts(self, result_file=None):
        """ Gait bout classification using acceleration data in the vertical direction from the lumbar location.

        Parameters
        ----------
        result_file : str
            Optional argument that accepts .h5 filepath string to save resulting predictions to.
            None by default. (ie. myfolder/myfile.h5)

        """
        import pickle
        import pandas as pd
        import os
        import deepdish as dd
        import util

        print('\tClassifying bouts of gait...')

        # Load model and feature order
        model_filename = os.path.dirname(os.path.realpath(__file__)) + '/model/model.pkl'
        features_filename = os.path.dirname(os.path.realpath(__file__)) + '/model/feature_order.txt'
        model = pickle.load(open(model_filename, 'rb'))
        feature_order = open(features_filename, 'r').read().splitlines()
        model_sample_rate = 50.

        # Load data and convert to g
        raw_y_accel, ts = util._load_data(self, self.down_sample)
        y_accel = raw_y_accel / 9.80665

        # Resample data if necessary
        if self.down_sample > model_sample_rate:
            data, timestamps = util._resample_data(y_accel, ts, str(1000./model_sample_rate) + 'ms')
        elif self.down_sample == model_sample_rate:
            data = pd.DataFrame({'y': y_accel})
            timestamps = pd.DatetimeIndex(ts.astype('datetime64[ms]'))
        elif self.down_sample < model_sample_rate:
            print('Data sample rate too low for bout detection model. Minimum sample rate required: ' +
                            str(model_sample_rate) + ' hz, aborting...')
            return

        # Extract signal features from vertical acceleration data
        feature_set, start_times_list, end_times_list = util._extract_signal_features(data, timestamps, model_sample_rate)
        feature_set = feature_set[feature_order]

        # Predict
        try:
            pred = model.predict(feature_set)
            predictions_df = pd.DataFrame(
                {'prediction': pred, 'window_start_time': start_times_list, 'window_end_time': end_times_list})
        except:
            print('Unable to make predictions from signal features, aborting...')
            return

        # Save predictions to hdf file
        if result_file:
            try:
                if not result_file.endswith('.h5'):
                    result_file += '.h5'
                predictions_dict = {}
                predictions_dict['predictions'] = predictions_df

                dd.io.save(result_file, predictions_dict)
            except:
                print('Unable to save data: Please make sure your results directory exists, aborting...')
                return

        print('\tBout classification complete!\n')

        return predictions_df



import setuptools

def requirements():
    with open('requirements.txt', "r") as fh:
        return [x for x in fh.read().split('\n') if x]

setuptools.setup(name='gaitpy',
                 version='1.1',
                 description='Read and process raw vertical accelerometry data from a lumbar sensor during gait; calculate clinical gait characteristics.',
                 url='http://github.com/PfizerRD/gaitpy',
                 packages=setuptools.find_packages(),
                 author='Matthew Czech',
                 author_email='czech1117@gmail.com',
                 keywords=['gait', 'lumbar', 'sensor', 'digital', 'wearable', 'python', 'inverted pendulum', 'czech'],
                 classifiers=["Programming Language :: Python :: 2.7",
                              "License :: OSI Approved :: MIT License"],
                 license='MIT',
                 zip_safe=False,
                 install_requires=requirements(),
                 include_package_data=True)


def _butter_lowpass(cutoff, fs, order):
    from scipy.signal import butter
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    return b, a

def _butter_lowpass_filter(data, fs, cutoff=20, order=4):
    from scipy import signal
    b, a = _butter_lowpass(cutoff, fs, order=order)
    filtered_signal = signal.filtfilt(b, a, data)
    return filtered_signal

def _band_pass_filter(data_df, sampling_rate, bp_cutoff, order, channels=['X', 'Y', 'Z']):
    from scipy import signal
    import pandas as pd
    import warnings
    import numpy as np

    data = data_df[channels].values

    # Calculate the critical frequency (radians/sample) based on cutoff frequency (Hz) and sampling rate (Hz)
    critical_frequency = [bp_cutoff[0]* 2.0 / sampling_rate, bp_cutoff[1]* 2.0 / sampling_rate]

    # Get the numerator (b) and denominator (a) of the IIR filter
    [b, a] = signal.butter(N=order, Wn=critical_frequency, btype='bandpass', analog=False)

    # Apply filter to raw data
    if np.isnan(np.sum(data)):
        data = np.nan_to_num(data)
        warnings.warn('There are NaN values in your acceleration data. Converting them to 0...')
    bp_filtered_data = signal.filtfilt(b, a, data, padlen=10, axis=0)

    new_channel_labels = [ax + '_bp_filt_' + str(bp_cutoff) for ax in channels]

    data_df[new_channel_labels] = pd.DataFrame(bp_filtered_data)

    return data_df


def _detect_peaks(y, prominence):
    from scipy.signal import find_peaks
    peaks, properties = find_peaks(y, prominence=prominence)

    return peaks

def _cwt(y_accel, sample_rate, ic_prom, fc_prom):
    from scipy import signal, integrate
    import pywt
    import pandas as pd

    # CWT wavelet scale parameters
    scale_cwt1 = float(sample_rate) / 5.0
    scale_cwt2 = float(sample_rate) / 6.0

    # Detrend data
    detrended_data = signal.detrend(y_accel)

    # Low pass filter if less than 40 hz
    if sample_rate >= 40:
        filtered_data = _butter_lowpass_filter(detrended_data, sample_rate)
    elif sample_rate < 40:
        filtered_data = detrended_data

    # cumulative trapezoidal integration
    integrated_data = integrate.cumtrapz(-filtered_data)

    # Gaussian continuous wavelet transform
    cwt_1, freqs = pywt.cwt(integrated_data, scale_cwt1, 'gaus1')
    differentiated_data = cwt_1[0]

    # initial contact (heel strike) peak detection
    ic_peaks = _detect_peaks(pd.Series(-differentiated_data), ic_prom)

    # Gaussian continuous wavelet transform
    cwt_2, freqs = pywt.cwt(-differentiated_data, scale_cwt2, 'gaus1')
    re_differentiated_data = cwt_2[0]

    # final contact (toe off) peak detection
    fc_peaks = _detect_peaks(pd.Series(re_differentiated_data), fc_prom)

    return ic_peaks, fc_peaks

def _calculate_sensor_height(subject_height, units, sensor_height_ratio):
    import warnings
    # calculate lumbar sensor height in meters
    if type(subject_height) is int or type(subject_height) is float:
        if units == 'inches':
            subj_height = subject_height * 0.0254
        elif units == 'centimeter':
            subj_height = subject_height * 0.01
        elif units == 'meter':
            subj_height = subject_height
        else:
            raise ValueError('Unable to identify units for subject height: Please make sure subject_height_units is correctly set.')

        sensor_height = subj_height * sensor_height_ratio
    else:
        warnings.warn('Subject height must be provided as an integer or float. Without subject height, gaitpy will not be able to calculate spatial features.')
        sensor_height = None

    return sensor_height

def _cwt_feature_extraction(cwt_features, sensor_height):
    import numpy as np

    # Calculate CWT model features
    # steps
    cwt_features['steps'] = len(cwt_features)

    # gait cycle duration
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'stride_duration'] = (cwt_features.IC.shift(-2) - cwt_features.IC) / 1000
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'stride_duration_asymmetry'] = abs(cwt_features.stride_duration.shift(-1) - cwt_features.stride_duration)

    # step time
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'step_duration'] = (cwt_features.IC.shift(-1) - cwt_features.IC) / 1000
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'step_duration_asymmetry'] = abs(cwt_features.step_duration.shift(-1) - cwt_features.step_duration)

    # cadence
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'cadence'] = 60 / cwt_features.step_duration

    # initial double support
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'initial_double_support'] = (cwt_features.FC_opp_foot - cwt_features.IC) / 1000
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'initial_double_support_asymmetry'] = abs(cwt_features.initial_double_support.shift(-1) - cwt_features.initial_double_support)

    # terminal double support
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'terminal_double_support'] = (cwt_features.FC - cwt_features.IC.shift(-1)) / 1000
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'terminal_double_support_asymmetry'] = abs(cwt_features.terminal_double_support.shift(-1) - cwt_features.terminal_double_support)

    # double support
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'double_support'] = cwt_features.initial_double_support + cwt_features.terminal_double_support
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'double_support_asymmetry'] = abs(cwt_features.double_support.shift(-1) - cwt_features.double_support)

    # single limb support
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'single_limb_support'] = (cwt_features.IC.shift(-1) - cwt_features.FC.shift(1)) / 1000
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'single_limb_support_asymmetry'] = abs(cwt_features.single_limb_support.shift(-1) - cwt_features.single_limb_support)

    # stance
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'stance'] = (cwt_features.FC - cwt_features.IC) / 1000
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'stance_asymmetry'] = abs(cwt_features.stance.shift(-1) - cwt_features.stance)

    # swing
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'swing'] = (cwt_features.IC.shift(-2) - cwt_features.FC) / 1000
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'swing_asymmetry'] = abs(cwt_features.swing.shift(-1) - cwt_features.swing)

    # step length
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'step_length'] = 2*(np.sqrt((2*sensor_height*cwt_features.CoM_height)-(cwt_features.CoM_height**2)))
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'step_length_asymmetry'] = abs(cwt_features.step_length.shift(-1) - cwt_features.step_length)

    # stride length
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'stride_length'] = cwt_features.step_length.shift(-1) + cwt_features.step_length
    # asymmetry
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'stride_length_asymmetry'] = abs(cwt_features.stride_length.shift(-1) - cwt_features.stride_length)

    # step velocity
    cwt_features.loc[cwt_features['Gait_Cycle'] == 1, 'gait_speed'] = cwt_features.stride_length / cwt_features.stride_duration

    return(cwt_features)

def _optimization(timestamps, ic_peaks, fc_peaks):
    import pandas as pd
    import numpy as np

    # Parameters
    gait_cycle_forward_ic = 2.25  # maximum allowable time (seconds) between initial contact of same foot
    loading_forward_fc = gait_cycle_forward_ic * 0.2  # maximum time (seconds) for loading phase
    stance_forward_fc = (gait_cycle_forward_ic / 2) + loading_forward_fc  # maximum time (seconds) for stance phase

    # Optimization 1: ---Loading Response--- Each IC requires 1 forward FC within 0.225 seconds (opposite foot toe off)
    #              2: ---Stance Phase--- Each IC requires atleast 2 forward FC's within 1.35 second (2nd FC is current IC's matching FC)
    ic_times = timestamps.iloc[ic_peaks]
    fc_times = timestamps.iloc[fc_peaks]

    optimized_gait = pd.DataFrame([])
    for i in range(0, len(ic_times)):
        current_ic = ic_times.iloc[i]
        loading_forward_max = current_ic + (loading_forward_fc * 1000.)
        stance_forward_max = current_ic + (stance_forward_fc * 1000.)

        loading_forward_fcs = fc_times[(fc_times > current_ic) & (fc_times < loading_forward_max)]
        stance_forward_fcs = fc_times[(fc_times > current_ic) & (fc_times < stance_forward_max)]
        stance_forward_ics = ic_times[((ic_times > current_ic) & (ic_times < stance_forward_max))]

        if len(loading_forward_fcs) == 1 and len(stance_forward_fcs) >= 2:
            icfc = pd.DataFrame({'IC': [current_ic], 'FC': [stance_forward_fcs.iloc[1]], 'FC_opp_foot': [stance_forward_fcs.iloc[0]]},
                                columns=['IC', 'FC', 'FC_opp_foot'])
            optimized_gait = optimized_gait.append(icfc)

    optimized_gait = optimized_gait.reset_index(drop=True)
    optimized_gait['CoM_height'] = np.nan
    optimized_gait['Gait_Cycle'] = 0

    # Optimization 3: ---Gait Cycles--- Each ic requires atleast 2 ics within 2.25 seconds after
    for i in range(0, len(optimized_gait) - 2):
        current_ic = optimized_gait.IC.iloc[i] / 1000.
        post_ic = optimized_gait.IC.iloc[i + 1] / 1000.
        post_2_ic = optimized_gait.IC.iloc[i + 2] / 1000.

        interval_1 = abs(post_ic - current_ic)
        interval_2 = abs(post_2_ic - current_ic)

        if interval_1 <= gait_cycle_forward_ic and interval_2 <= gait_cycle_forward_ic:
            optimized_gait.Gait_Cycle.iloc[i] = 1

    return optimized_gait

def _height_change_com(optimized_gait, timestamps, gait_data, sample_rate):
    from scipy import signal, integrate
    # Changes in Height of the Center of Mass
    for i in range(0, len(optimized_gait) - 1):
        ic_index = timestamps.index[timestamps == optimized_gait.IC[i]].item()
        post_ic_index = timestamps.index[timestamps == optimized_gait.IC[i + 1]].item()

        step_raw = gait_data.loc[ic_index:post_ic_index]
        step_detrended = signal.detrend(step_raw)
        if sample_rate >= 40:
            step_filtered = _butter_lowpass_filter(step_detrended, sample_rate)
        elif sample_rate < 40:
            step_filtered = step_detrended
        step_integrate_1 = integrate.cumtrapz(step_filtered) / sample_rate
        step_integrate_2 = integrate.cumtrapz(step_integrate_1) / sample_rate

        h = max(step_integrate_2) - min(step_integrate_2)
        optimized_gait.CoM_height.iloc[i] = h

    return optimized_gait

def _resample_data(y_accel, timestamps, new_fs):
    import pandas as pd

    # Concatenate data and timestamps
    data = pd.DataFrame({'ts': timestamps.astype('datetime64[ms]'), 'y': y_accel})
    data.set_index('ts', inplace=True)

    # Resample data
    resampled_data = pd.DataFrame(data['y'].resample(new_fs).mean())

    # Create resampled timestamp dataframe
    resampled_timestamps = resampled_data.index

    # Reset index of resampled data
    resampled_data.reset_index(drop=True, inplace=True)

    return resampled_data, resampled_timestamps

def _load_data(self, down_sample):
    import pandas as pd
    import numpy as np

    # Load data
    try:
        if type(self.data) is str:
            data_df = pd.read_csv(self.data)
        elif type(self.data) is pd.core.frame.DataFrame:
            data_df = self.data
        else:
            raise Exception('Unable to load data: Please make sure the data is in the correct format')
    except:
        raise Exception('Unable to load data: Please make sure you have provided the correct filepath.')

    # Convert timestamps to milliseconds. Convert data to m/s^2. Flip data if specified.
    try:
        if self.ts_units == 's':
            timestamps = data_df[self.ts_col_name] * 1000.
        elif self.ts_units == 'ms':
            timestamps = data_df[self.ts_col_name]
        elif self.ts_units == 'us':
            timestamps = data_df[self.ts_col_name] / 1000.

        if self.v_acc_units == 'm/s^2':
            y_accel = data_df[self.v_acc_col_name]
        elif self.v_acc_units == 'g':
            y_accel = data_df[self.v_acc_col_name] * 9.80665
        else:
            raise Exception("Unable to load data: Please make sure the units you provide are either 'm/s^2' or 'g'")

        if self.flip == True:
            y_accel = y_accel * -1
    except:
        raise Exception('Unable to load data: Please make sure your columns are labeled correctly.')

    # Downsample data
    if self.sample_rate > down_sample:
        resampled_data, resampled_ts = _resample_data(y_accel, timestamps, str(1000. / down_sample) + 'ms')
        data = pd.Series(resampled_data.y)
        ts = pd.Series(resampled_ts.astype(np.int64).values // 10**6)
    elif self.sample_rate == down_sample:
        data = y_accel
        ts = timestamps
    elif self.sample_rate < down_sample:
        print('Data sample rate too low for extract_feature analysis. Minimum sample rate required: ' +
              str(down_sample) + ' hz, aborting...')
        return

    return data, ts

def _extract_signal_features(data, timestamps, sample_rate, window_length=3.0):
    import pandas as pd
    import signal_features as sf
    import warnings

    data.reset_index(drop=True, inplace=True)
    filtered_data_df = _band_pass_filter(data, sample_rate, [0.5, 3.0], 1, channels=['y'])
    total_data_channels = ['y_bp_filt_[0.5, 3.0]']

    # Initialize final DataFrame
    location_feature_and_label_set = pd.DataFrame()

    # Segment into 3 second windows
    total_samples = filtered_data_df.shape[0]
    window_samples = sample_rate * window_length
    total_windows = round(total_samples / float(window_samples))

    start_times_list = []
    end_times_list = []

    for win in range(int(total_windows)):
        # Isolate data into windows
        current_win_start = int(window_samples * win + 1)
        current_win_end = int(current_win_start + window_samples - 1)
        if current_win_end >= filtered_data_df.shape[0]:
            window_data_df = filtered_data_df.loc[current_win_start:, :]
            end_time = timestamps[-1]
        else:
            window_data_df = filtered_data_df.loc[current_win_start:current_win_end, :]
            end_time = timestamps[current_win_end]
        start_time = timestamps[current_win_start]
        window_data_df.reset_index(drop=True, inplace=True)

        # Extract Features
        try:
            features_df = sf._signal_features(window_data_df, total_data_channels, sample_rate)
        except:
            warnings.warn('Error calculating signal features for 3-second window between '+str(current_win_start)+' and '+str(current_win_end)+'.')
            continue

        # Discard window if NaN's in feature matrix
        if features_df.isnull().values.any():
            continue

        # Rename columns with device location
        location_columns = ['lumbar_' + s for s in features_df.columns]
        features_df.columns = location_columns

        # Aggregate features for each window
        location_feature_and_label_set = location_feature_and_label_set.append(features_df, ignore_index=True)
        start_times_list.append(start_time)
        end_times_list.append(end_time)

    return location_feature_and_label_set, start_times_list, end_times_list

def _concatenate_windows(windows, window_length):
    import pandas as pd

    concatenated_windows_df = pd.DataFrame([])
    windows = windows.reset_index(drop=True)
    window = 0
    while window < len(windows):
        # concatenate windows of gait from bout windows
        start_time = windows.window_start_time[window]
        window_boolean = True
        while window_boolean:
            end_time = windows.window_end_time[window]
            if window == (len(windows) - 1):
                window_boolean = False
            elif (windows.window_start_time[window+1] - windows.window_start_time[window]).total_seconds() == float(window_length):
                window = window + 1
            else:
                window_boolean = False
        window = window + 1

        # append concatenated windows start and end time to dataframe
        start_end_df = pd.DataFrame(data={'start_time': [start_time],
                                          'end_time': [end_time],
                                          'bout_length': [(end_time - start_time).total_seconds()]})
        concatenated_windows_df = concatenated_windows_df.append(start_end_df)
        concatenated_windows_df.reset_index(drop=True, inplace=True)

    return concatenated_windows_df
